#pragma once

#include <bt/node_task.hpp>

namespace bt {
using yield = std::suspend_always;

template <typename Context> class behavior_node {
public:
  virtual node_task tick(Context &ctx) noexcept = 0;

  virtual ~behavior_node() = default;
};

template <typename Context>
class composite_node : public behavior_node<Context> {
public:
  void add_child(std::unique_ptr<behavior_node<Context>> &&child) {
    _children.push_back(std::move(child));
  }

protected:
  std::vector<std::unique_ptr<behavior_node<Context>>> _children;
};

template <typename Context>
class decorator_node : public behavior_node<Context> {
public:
  decorator_node(std::unique_ptr<behavior_node<Context>> child)
      : _child(std::move(child)) {}
  
  void set_child(std::unique_ptr<behavior_node<Context>> &&child) {
    _child = std::move(child);
  }

protected:
  std::unique_ptr<behavior_node<Context>> _child;
};

template <typename Context>
class sequence_node : public composite_node<Context> {
public:
  node_task tick(Context &ctx) noexcept override {
    std::size_t idx = 0;
    while (idx < this->_children.size()) {
      auto task = this->_children[idx]->tick(ctx);
      while (!task.done()) {
        task.resume();

        if (idx < this->_children.size() - 1)
          co_await yield{};
      }

      if (task.result() == task_result::failure) {
        co_return task_result::failure;
      }

      ++idx;
    }

    co_return task_result::success;
  }
};

template <typename Context>
class selector_node : public composite_node<Context> {
public:
  node_task tick(Context &ctx) noexcept override {
    for (std::size_t idx = 0; idx < this->_children.size(); ++idx) {
      auto task = this->_children[idx]->tick(ctx);
      while (!task.done()) {
        task.resume();
        co_await yield{};
      }

      if (task.result() == task_result::success) {
        co_return task_result::success;
      }
    }
    co_return task_result::failure;
  }
};

template <typename Context>
class parallel_node : public composite_node<Context> {
public:
  node_task tick(Context &ctx) noexcept override {
    std::vector<node_task> tasks;
    for (auto& child : this->_children) {
      tasks.push_back(child->tick(ctx));
    }

    while (!tasks.empty()) {
      for (auto& task : tasks) {
        if (task.done())
          continue;
        
        task.resume();
        co_await yield{};
      }

      std::erase_if(tasks, [](auto& task) { return task.done(); });
    }

    co_return task_result::success;
  }
};


template <typename Context, typename TickFn>
class action_node : public behavior_node<Context> {
public:
  action_node(TickFn tick) noexcept : _tick(tick) {}

  node_task tick(Context &ctx) noexcept override { return _tick(ctx); }

private:
  TickFn _tick;
};

template <typename Context, typename ConditionFn>
class conditional_node : public behavior_node<Context> {
public:
  conditional_node(ConditionFn condition) noexcept : _condition(condition) {}

  node_task tick(Context &ctx) noexcept override {
    if (_condition(ctx)) {
      co_return task_result::success;
    }
    co_return task_result::failure;
  }

private:
  ConditionFn _condition;
};

template <typename Context>
class inverter_node : public decorator_node<Context> {
public:
  inverter_node(std::unique_ptr<behavior_node<Context>> child)
      : decorator_node<Context>(std::move(child)) {}
      
  node_task tick(Context &ctx) noexcept override {
    auto task = this->_child->tick(ctx);
    while (!task.done()) {
      task.resume();
      co_await yield{};
    }

    if (task.result() == task_result::success) {
      co_return task_result::failure;
    }
    co_return task_result::success;
  }
};
}